package se.umu.jobsubmission.impl.broker; 

import java.util.List;
import java.util.Vector;

import org.apache.axis.message.MessageElement;
import org.apache.log4j.Logger;
import org.globus.gsi.GlobusCredential;
import org.ggf.schemas.jsdl._2005._11.jsdl.Application_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDescription_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl.Resources_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl.CPUArchitecture_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl.RangeValue_Type;
import org.ggf.schemas.jsdl._2005._11.jsdl_posix.POSIXApplication_Type;
import org.globus.mds.glue.GLUECERPType;
import org.globus.wsrf.encoding.ObjectDeserializer;

import se.umu.jobsubmission.InvalidJobDescriptionFaultType;
import se.umu.jobsubmission.JobObjectiveType;
import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.util.ReservationUtils;

import org.globus.wsrf.utils.FaultHelper; 

// TODO: Check "policy limits" (such as GLUE:X:Y:maxCount) 
// in Broker.filter() 

public class Broker {

    private static Logger logger = Logger.getLogger(Broker.class); 

    // Estimates job start time, job execution time
    // and job file staging time. 
    Predictor startTimePredictor; 
    
    // TODO: Make this more general, e.g. by using a list of selectors
    // and the chain of responsibility pattern...
    Selector earliestStartSelector;  
    Selector earliestCompletionSelector; 
    
    private ReservationUtils reservationUtils; 

    public Broker(int numberOfThreads) {
    	this.reservationUtils = null; 
    	this.startTimePredictor = new StartTimePredictor(); 
    	
    	this.earliestStartSelector = new EarliestStartSelector(
    			startTimePredictor, numberOfThreads); 
    	
    	this.earliestCompletionSelector = new EarliestCompletionSelector(
    			startTimePredictor, numberOfThreads); 
    }

    // Verify that the jobDescriptions are ok
    // Ok means: contains an Application element
    // which in terms contains an POSIXApplication
    // element specifying an Executable.
    // 
    public void verify(JobDefinition_Type jobDesc) 
    throws InvalidJobDescriptionFaultType {
 
    	POSIXApplication_Type posix = null; 
    	try {
    		Application_Type application =  jobDesc.getJobDescription().
					getApplication();
    		MessageElement any[] = application.get_any(); 
    		posix = (POSIXApplication_Type) ObjectDeserializer.toObject(any[0], 
    				POSIXApplication_Type.class); 
    	} catch (Exception e) {
    		InvalidJobDescriptionFaultType f = new InvalidJobDescriptionFaultType();            
			FaultHelper helper = new FaultHelper(f); 
			helper.addDescription("Invalid job description, reason unknown");     				
			throw (InvalidJobDescriptionFaultType) helper.getFault();       		
    	}
    		
    	if(posix.getExecutable() == null) {
    		InvalidJobDescriptionFaultType f = new InvalidJobDescriptionFaultType();            
    		FaultHelper helper = new FaultHelper(f); 
    		helper.addDescription("Invalid job description: " 
					+ "executable must be specified"); 
    		throw (InvalidJobDescriptionFaultType) helper.getFault();     
    	}
    }


    // Removes all resources that fails to fulfill requirements
    // for hardware etc. from the list of resources. 
    // Returns the other resources. 
    public List filter(List resources, JobDefinition_Type jobDesc) {

        List approvedResources = new Vector(); 

        // For each of the attributes 
        // ProcessCount, CPUTime, DiskSpace, PhysicalMemory
        // and CPUArchitecture that jobDescription 
        // specifies, ensure that the resource match 
        // the attribute. 

        logger.debug("Filtering resources...."); 
        
        // Ensure that job description is ok. 
        if(jobDesc == null) {
        	logger.error("No job description specified"); 
        	return null;     
        }
        JobDescription_Type desc = jobDesc.getJobDescription(); 
        if(desc == null) {
        	logger.error("No job description specified"); 
        	return null; 
        }
        
        // User has no requirements what so ever, all resources all eligible
        Resources_Type resource = desc.getResources(); 
        if(resource == null) {
        	logger.info("No requirements specified, accepting all resources"); 
        	return resources; 
        }
        
        String jobArchitecture = null; 
        CPUArchitecture_Type architecture = resource.getCPUArchitecture(); 
        if(architecture != null) {
        	jobArchitecture = architecture.getCPUArchitectureName().getValue(); 
        }
       
        RangeValue_Type memory = resource.getIndividualPhysicalMemory(); 
        int jobMemory = -1; 
        if(memory != null) {
        	jobMemory = (int) memory.getExact()[0].get_value();
        	/*
            String memoryString = ((Text)jobMemory.getContent().get(0)).getTextTrim(); 
            jobMemoryInt = Integer.parseInt(memory); */
        }
        else
            logger.debug("User did not specify \"PhysicalMemory\""); 

        RangeValue_Type count = resource.getTotalCPUCount(); 
        int jobCount = -1; 
        if(count != null) {
        	jobCount = (int) count.getExact()[0].get_value(); 
        } 
        else {
            logger.debug("User did not specify \"ProcessCount\""); 
        }

        RangeValue_Type diskSpace = resource.getIndividualDiskSpace(); 
        int jobDiskSpace = -1; 
        if(diskSpace != null) {
        	jobDiskSpace = (int) diskSpace.getExact()[0].get_value(); 
        }
        else {
            logger.debug("User did not specify \"DiskSpace\""); 
        }
        
        // Test all found resources against these requirements. 
        for(int i = 0; i < resources.size(); i++) {

            GLUECERPType resourceInfo = (GLUECERPType) resources.get(i); 

            String resourceName = resourceInfo.getCluster(0).getSubCluster(0).
				getHost(0).getName(); 

            // CPUArchitecture
            // TODO: Might want to consider operator in JSDL 
            // but does (architecture != sparc) really make sense? 
            // --> operators removed from JSDL Spec. 
            if(jobArchitecture != null) {
                String resourceArchitecture = resourceInfo.getCluster(0).
                    getSubCluster(0).getHost(0).getArchitecture().getPlatformType(); 
                if(resourceArchitecture == null) {
                    logger.debug("Resource " + resourceName + 
                            " does not specify architecture, rejecting it"); 
                    continue; 
                }
                
                // TODO, There exist some constants in the generated
                // JSDL types? Use these!

                // This is a mess!
                // ARC (and possible, others) 
                // allows free text format for architecture...
                
               // Does GLUE specify architecture constants? 
               // In that case, use them.
               // Make translation of architecture constants in 
               // *InformationConverter to some standard format, 
               // can not use ARC-specific strings in here!
               if(jobArchitecture.equals("x86") &&
                       !resourceArchitecture.equals("i686")) {
                   logger.debug("Resource " + resourceName + 
                           " rejected due to wrong architecture"); 
                   continue; 
               }
               
               // Other defined values (in JSDL) 
               // sparc, powerpc, parisc, mips, ia86, arm, other
               // Notably missing: alpha
               
               // TODO: Add tests for other architectures...

            }
            
            if(jobMemory >= 0) {
                long resourceMemory = resourceInfo.getCluster(0).
                    getSubCluster(0).getHost(0).getMainMemory().getRAMSize(); 
                if((long)jobMemory > resourceMemory) {
                    logger.debug("Resource " + resourceName + 
                            " rejected due to limited memory"); 
                    continue; 
                }
            }

            if(jobCount >=0) {
            	try {
            		int resourceTotalCount = Integer.parseInt(resourceInfo.
                        getComputingElement(0).getInfo().getTotalCPUs()); 
            		if(jobCount > resourceTotalCount) {
            			logger.debug("Resource " + resourceName + 
            			" rejected due to too few CPUs"); 
            			continue; 
            		}
            	} catch (NumberFormatException e) {
            		logger.error("Resource " + resourceName + 
            				" rejected due to incorrect CPU specification: " 
            				+ resourceInfo.getComputingElement(0).getInfo().
							getTotalCPUs()); 
            		
            	}
            }

            if(jobDiskSpace >= 0) {
                long resourceDiskSpace = resourceInfo.getCluster(0).
                    getSubCluster(0).getHost(0).getFileSystem(0).getAvailableSpace(); 
                if((long)jobDiskSpace > resourceDiskSpace) {
                    logger.debug("Resource " + resourceName + 
                            " rejected due to insufficient disk space"); 
                    continue; 
                }
            }
            
            logger.debug("Resource " + resourceName + " is approved for job."); 

            approvedResources.add(resourceInfo); 

        }

        return approvedResources; 

    }
    
    public void printResources(List resources) {
    	
    	GLUECERPType resource = null; 
    	
    	for(int i = 0; i < resources.size(); i++) {
    		resource = (GLUECERPType) resources.get(i); 
    		logger.debug("R: " + resource.getCluster(0).getSubCluster(0).
    				getHost(0).getName()); 
    	}
    	
    }
    
    /**
     * Returns the appropiate selector based on the objective
     * specified in the job preferences. 
     * 
     * @param preferences
     * @return Returns an appropiate selector. EarliestCompletionSelector
     * is returned if no (or an incorrect) selector is specified. 
     */
    public Selector getSelector(JobPreferencesType preferences) {    	
    	
    	Selector selector = null; 

        if(preferences == null) {
            return earliestCompletionSelector; 
        }
    	
    	try {
    		
    		JobObjectiveType objective = preferences.getJobRequirement().
				getJobObjective();    
    		if(objective.equals(JobObjectiveType.earliestJobCompletion))
    			selector = earliestCompletionSelector; 
    		else if(objective.equals(JobObjectiveType.earliestJobStart))
    			selector = earliestStartSelector; 
    		else {
    			logger.debug("Unknown job objective: " 
                                + objective.getValue());
    			selector = earliestCompletionSelector; 
    		}
    	} catch (Exception e) { 
    		logger.debug("No job objective specified, using" +
    		" earliest completion");
    		selector = earliestCompletionSelector;         
    	}
    	
    	return selector; 
    }

    /**
     * Selects the resources most apt to execute the job
     * by reordering the list of approved resources. 
     * 
     * @param approvedResources
     * @param jobDescription
     * @param preferences
     * @param selector
     * @return
     */
    public List select(List approvedResources, 
    		JobDefinition_Type jobDescription, 
    		JobPreferencesType preferences, 
			Selector selector, 
			GlobusCredential credential) {

        logger.debug("Selecting \"best\" resource"); 

        if(approvedResources == null || approvedResources.size() == 0) {
            logger.debug("No resources to select from!"); 
            return null;     
        }
        else {

        	logger.debug("Before reordering"); 
        	if(logger.isDebugEnabled())
        		printResources(approvedResources);
        	// Randomize list (gives better load balancing
        	// in case two resources are predicted to have equal
        	// performance). 
            int size = approvedResources.size(); 
            Object resourceArray[] = approvedResources.toArray(); 
            for(int i = 0; i < size; i++) {
            	int index = (int) (Math.random()*size); 
            	Object tmp = resourceArray[i]; 
            	resourceArray[i] = resourceArray[index]; 
            	resourceArray[index] = tmp; 
            }            
            approvedResources.clear(); 
            for(int i = 0; i < size; i++) {
            	approvedResources.add(resourceArray[i]); 
            }
        		
            // Let selector reorder resources. 
        	List result = selector.orderResources(approvedResources, 
        			jobDescription, preferences, credential);
        	
        	logger.debug("After reordering"); 
        	if(logger.isDebugEnabled())
        		printResources(result);
        	
        	return result; 
        }
 
    }
        
    // Returns the reservationUtils object. All methods for
    // managing reservations are moved to the ReservationUtils class. 
    public ReservationUtils getReservationUtils() {
    	return this.reservationUtils; 
    }
    
    public void setReservationUtils(ReservationUtils utils) {
    	this.reservationUtils = utils; 
    	((StartTimePredictor)this.startTimePredictor).setReservationUtils(utils); 
    }
    

}
