package se.umu.jobsubmission.impl.dispatcher; 

import java.net.URI;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.axis.MessageContext; 
import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.gsi.GlobusCredential;
import org.globus.mds.glue.GLUECERPType;
import org.globus.rsl.AbstractRslNode;
import org.globus.rsl.NameOpValue;
import org.globus.rsl.RslEvaluationException;
import org.globus.rsl.Value;
import org.nordugrid.job.gridftp.ARCGridFTPJob;
import org.nordugrid.job.gridftp.ARCGridFTPJobException;

import se.umu.jobsubmission.impl.datamanager.DataHelper;
import se.umu.jobsubmission.util.XRSLUtils;
import se.umu.reservation.ReservationInfo;
import se.umu.jobsubmission.util.SubmitterUtils; 


/*
   ARCDispatcher - submits jobs to ARCResources. 
   method requestJob takes an Object (contact of resource), 
   an jsdl job description (must be converted to xRSL before submission) 
   and an Object (reservation) 

   The xRSL job description is manipulated a bit to ensure that
   the GridFTP server handles is correctly. 

   */

public class ARCDispatcher implements Dispatcher {

    private static Logger logger = Logger.getLogger(ARCDispatcher.class); 

    private jsdl2xrsl descriptionTranslator; 

    public ARCDispatcher() {
        descriptionTranslator = new jsdl2xrsl(); 
    }
    
    /**
     * Removes the locally stored input files from the xRSL to 
     * avoid duplicated file staging. 
     * @param rsl
     * @param clientURL
     * @throws Exception
     */   
    public void removeStagedInputFiles(AbstractRslNode rsl) throws Exception {
    	
        NameOpValue inputfiles = rsl.getParam("inputfiles"); 

        // Values are stored as list (pair) of lists 
        if(inputfiles != null) {
        	
            List values = inputfiles.getValuesAsStrings(true); 
            Iterator iter = values.iterator(); 
            
            List removedFiles = new Vector(); 

            // Extract files stored on client host.
            // Return these files as a list of lists. 
            while(iter.hasNext()) {
                List l; 
                try {
                    l = (List) iter.next(); 
                }
                catch (ClassCastException e) {
                    throw new Exception("inputfiles must be a list of lists!"); 
                }
                
                // Format is (localURL remoteURL), e.g.:
                // N.B: localURL is local on the resource, 
                //      remoteURL is local to the client host. 
                String resourceName = (String) l.get(0); 
                String clientName = (String) l.get(1); 
                
                String scheme = null; 
                String path = null; 
                try {
                	URI clientURI = new URI(clientName); 
                	scheme = clientURI.getScheme(); 
                	path = clientURI.getPath(); 
                } catch (Exception e) {
                	logger.debug("Ignoring file " + clientName 
                			+ " due to malformed URL"); 
                	continue; 
                	// malformed URL... 
                }
                
                if("file".equals(scheme)) {
                	if(path == null || path.length() == 0) 
                		path = clientName.substring("file://".length()); 
 
                	// Add file to list of files that will be removed later
                	List value = new Vector(); 
                	value.add(new Value(resourceName)); 
                	value.add(new Value(clientName)); 
                	removedFiles.add(value); 
                }
            }
            
            // Now, remove all files that will be transferred by the
            // client 
            // by the dispatcher. 
            for(int i = 0; i < removedFiles.size(); i++) {
            	inputfiles.remove(removedFiles.get(i)); 
            }
            if(inputfiles.getValues().size() == 0) {
            	NameOpValue rel = rsl.removeParam("inputfiles");  
            	if(rel != null)
            		logger.debug("removed " + rel.getAttribute()); 
            }
            
        }
    }

    // Adds the rsl submitted to the user to the rsl 
    // can be used for resubmitting the job. 
    public void addUserRsl(AbstractRslNode rsl) throws Exception {

        AbstractRslNode userRsl; 
        try {
           userRsl = rsl.evaluate(); 
        } catch (RslEvaluationException e) {
            logger.error("Error copying rsl"); 
            throw e; 
        }

        NameOpValue clientxrsl = new NameOpValue("clientxrsl", 
                NameOpValue.EQ, userRsl.toRSL(true)); 
        rsl.add(clientxrsl); 

    }

    // If executable does not start with '/', it is a local file which
    // must be staged, and added to the "executables" relation
    public void fixExecutable(AbstractRslNode rsl) throws Exception {

        NameOpValue executableRelation = rsl.getParam("executable"); 
        if(executableRelation == null) 
            throw new Exception("xRSL must contain \"executable\" relation"); 

        List executableValues = executableRelation.getValuesAsStrings(true); 
        String executableValue = (String) executableValues.get(0); 


        // And now, for my final trick...
        // Replace executable with /bin/echo, and place the old executable first
        // in 'arguments' (the result of `echo arguments` 
        // executed on the remote host) 
        
        // Get argument relation
        NameOpValue argumentRelation = rsl.getParam("arguments"); 

        // No args, just add executable as only argument
        if(argumentRelation == null) {
            argumentRelation = new NameOpValue("arguments", 
                    NameOpValue.EQ, executableValue); 
            rsl.add(argumentRelation); 
        }
        // Arguments exist, add executable as first argument
        else {
            // Old arguments (type String) 
            List argumentValues = argumentRelation.getValuesAsStrings(true); 
            argumentRelation.clear(); 
            argumentRelation.add(executableValue); 
            Iterator iter = argumentValues.iterator(); 
            while(iter.hasNext()) 
                argumentRelation.add((String)iter.next()); 
            
        }

        executableRelation.clear(); 
        executableRelation.add("/bin/echo"); 
    }

    // Rename stdin to sstdin
    public void fixStdin(AbstractRslNode rsl) {

        NameOpValue stdinRelation = rsl.removeParam("stdin"); 
        if(stdinRelation == null) 
            return; 

        List stdinValues = stdinRelation.getValuesAsStrings(true); 
        String stdinValue = (String) stdinValues.get(0); 
        NameOpValue sstdinRelation = new NameOpValue("sstdin", NameOpValue.EQ,
               stdinValue);  
        rsl.add(sstdinRelation); 

    }

    // Add queue name (FIXME: only uses default queue ATM)
    // This should be fixed in ARCInformationFinder 
    // and ARCInformationConverter, not here...
    public void fixQueue(AbstractRslNode rsl, GLUECERPType resourceInfo) {

        String queueName = resourceInfo.getComputingElement(0).getName(); 

        NameOpValue queueRelation = new NameOpValue("queue", 
        		NameOpValue.EQ, 
                queueName); 
        rsl.add(queueRelation); 

    }
    
    // Adds default cputime (5 minutes) unless it is specified.     
    public void fixCpuTime(AbstractRslNode rsl) {
    	
    	NameOpValue cpuTimeRel = rsl.getParam("cputime"); 
    	if(cpuTimeRel == null) {
    		logger.debug("\"cputime\" not specified, assuming " 
    				+ Dispatcher.DEFAULT_CPU_TIME + " minutes"); 
    		cpuTimeRel = new NameOpValue("cputime", NameOpValue.EQ, 
    				"" + Dispatcher.DEFAULT_CPU_TIME); 
    		rsl.add(cpuTimeRel); 
    	}
    	
    }
 
    /**
     * Adds hostname of client submitting job to the RSL. 
     * @param rsl
     * @param clientURL
     */
    public void fixHostName(AbstractRslNode rsl, String clientURL) {
    	
    	logger.debug("Adding \"hostname\" = " + clientURL); 
    	NameOpValue hostNameRelation = new NameOpValue("hostname", 
    			NameOpValue.EQ, 
				clientURL); 
    	rsl.add(hostNameRelation); 			
    	
    }
    
    // Add name of client software used to submit job. 
    public void fixClientSoftware(AbstractRslNode rsl) {    	
    	NameOpValue clientSoftwareRelation = new NameOpValue("clientsoftware", 
    			NameOpValue.EQ, 
				"JobSubmissionService-0.1"); 
    	rsl.add(clientSoftwareRelation);     	
    }

    // Performs the required manipulations to the rsl. 
    // Add a copy of the submitted rsl in the clientxrsl relation, 
    // adds (action=request) and (savestate=yes) to rsl, 
    // add files to be staged in/out to inputfiles/outputfiles
    // renames (stdin=*) to (sstdin=*). 
    // Moved command from executable to arguments and set
    // executable to /bin/echo
    public void prepareRsl(AbstractRslNode rsl, 
    		GLUECERPType resourceInfo, 
			String clientURL) throws Exception {

        addUserRsl(rsl); 

        NameOpValue action = new NameOpValue("action", 
        		NameOpValue.EQ, 
				"request"); 
        rsl.add(action); 
        NameOpValue savestate =  new NameOpValue("savestate", 
        		NameOpValue.EQ, 
				"yes"); 
        rsl.add(savestate); 

        // Add relations (if not already defined), 
        // to ensure that files are staged out. 
        XRSLUtils.addIOStaging(rsl, "stdout", "outputfiles"); 
        XRSLUtils.addIOStaging(rsl, "stderr", "outputfiles"); 
        XRSLUtils.addIOStaging(rsl, "gmlog", "outputfiles"); 
 
        fixStdin(rsl); 
        fixExecutable(rsl); 
        fixQueue(rsl, resourceInfo); 
        fixCpuTime(rsl); 
        fixHostName(rsl, clientURL);
        fixClientSoftware(rsl); 

    }
    
    /* 
     * Resource is a GLUECERPType
     * jobDescription is a JobDefinition_Type describing the job in JSDL
     * reservation contains the RPs from the AgreementResource. 
     */
    public Object requestJob(Object resource,    
                             Object jobDescription, 
                             Object reservation, 
							 String clientURL, 
							 boolean clientStagesInput, 
							 GlobusCredential credential) throws Exception {

        GLUECERPType resourceInfo; 
        try {
            resourceInfo = (GLUECERPType) resource; 
        } catch (ClassCastException e) {
            logger.debug("Resource description of incorrect type"); 
            throw new Exception("Resource description of incorrect type", e); 
        }

        // ng-cluster-contactstring is now stored as 
        // Cluster:SubCluster:Host:Name to ensure conformance with GT4... 
        // String resourceURI = resourceInfo.getCluster(0).getUniqueID();
        String resourceURI = resourceInfo.getCluster(0).getSubCluster(0).getHost(0).getName(); 

        ARCGridFTPJob job = null; 
      
        try {
        	
            job = new ARCGridFTPJob(resourceURI);
            
            // Set user's delegated credential as proxy used by job. 
            job.addProxy(credential); 
            
    		// Connect to resource (using user's delegated proxy)
			job.Connect(); 
			
        } catch (ARCGridFTPJobException e) {
            logger.error("Could not create ARCGridFTPJob: " + e.getMessage()); 
            throw e; 
        }
        
        AbstractRslNode rslNode = (AbstractRslNode) 
			descriptionTranslator.translate(jobDescription); 

        if(rslNode == null) {
            logger.error("Could not translate job description"); 
            throw new ARCGridFTPJobException
                ("could not translate job description"); 
        }
        
        // Add some commands to rsl, and 
        // add all input/output files to lists of files to stage. 
        prepareRsl(rslNode, resourceInfo, clientURL); 
        
        // Add reference to local reservation if it exists. 
        try {
        	if(reservation != null) {
        		ReservationInfo resInfo = (ReservationInfo) reservation; 
        		String localID = resInfo.getReservation().getLocalID(); 
        		
        		logger.debug("Adding (advres = " + localID + ")"); 
        		if(localID != null) {        		
        			NameOpValue advres = new NameOpValue(
        					"advres", 
							NameOpValue.EQ, 
							localID); 
        			rslNode.add(advres); 
        		}
        	}
        } catch (Exception e) {        	
        	logger.error("Error in reservation, ignoring it"); 
        }
        
        if(clientStagesInput) {
        	this.removeStagedInputFiles(rslNode); 
        }

        String rslString = rslNode.toRSL(true); 
        logger.debug("Final xRSL is: \n" + rslString);        
        
        // Submit job
        String resourceContact = null; 
        try {
            job.Submit(rslString);
            resourceContact = job.Id(); 
            logger.debug("Local id is " + resourceContact);              
            
            if(!clientStagesInput) {             
            	
            	// Use DataHelper to identify and stage input files
            	// We need to know the job ID to stage the files, 
            	// Hence, this code can not be moved to the ARCDataManager. 
            	try { 
            		
            		List localInputFiles = DataHelper.getLocalInputFiles(
            				(JobDefinition_Type)jobDescription); 
            		if(localInputFiles != null) {
            			logger.info("Staging input files"); 
            			DataHelper.stageLocalInputFiles(job.Id(), 
            					localInputFiles, 
            					clientURL, 
								credential); 
            		}
            	} catch (Exception e) {
            		job.Cancel(); 
            		job.Disconnect(); 
            		throw new ARCGridFTPJobException("Error staging files: " + 
            				e.getMessage()); 
            	}
            }
            else {
            	logger.info("No file staging performed (client handles it)");             
            }
            job.Disconnect();
        }
        catch (ARCGridFTPJobException e) {
            logger.error("Job submission failed: " + e.getMessage()); 
            throw e; 
        }

        // Generate globalid, which is resourceURL/jobid
        String globalId = resourceURI; 
        if(globalId.charAt(globalId.length() -1) != '/')
            globalId += "/"; 
        globalId += resourceContact; 

        return globalId; 
    }

}
