package se.umu.jobsubmission.client; 

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.net.InetAddress;
import java.util.List;
import java.util.Vector;

import javax.xml.rpc.Stub;

import org.apache.axis.AxisFault;
import org.apache.axis.message.addressing.Address;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.axis.util.Util;
import org.globus.wsrf.client.ServiceURL;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.utils.FaultHelper;
import org.oasis.wsrf.faults.BaseFaultType;
import org.xml.sax.InputSource;

import se.umu.jobsubmission.ClientConfiguration;
import se.umu.jobsubmission.InvalidJobDescriptionFaultType;
import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequest;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.JobSubmission;
import se.umu.jobsubmission.JobSubmissionServiceAddressingLocator;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.NoResourcesFoundFaultType;
import se.umu.jobsubmission.impl.jobsubmissionservice.JobSubmissionConstants;
import se.umu.jobsubmission.util.ClientUtils;
import se.umu.jobsubmission.util.JobDescriptionTranslator;


/**
 * Simple client for job submission. 
 * The client accepts four parameters, 
 * the job description (mandatory),
 * a job preferences document (optional), 
 * a list of index servers (optional), 
 * and an URL to the JSS to use (optional)
 */

public class SubmitJob {

    static {
        Util.registerTransport(); 
    }

    private String jobDescriptionFile; 
    private String jobPreferencesFile; 
    private String indexServers[]; 
    private String URI; 
    private String host; 
    private String clientHost; 
    private boolean clientStagesInput; 
    
    private String[] converterClasses; 
    private String[] stagingTargetResolverClasses; 
    private String clientConfigFile; 
    private String clientSecurityDescriptorFile; 
    
    private JobDescriptionTranslator jobDescriptionTranslators[]; 

    private static final String optionsDesc = 
        "OPTIONS:\n" +
        "\n"+
        "Help:\n" +
        " -help                 display help.\n" +
        " -usage                display usage.\n" +
        "Usage:\n" + 
        " -f                    Specify file that contains job description\n" + 
        " -i                    Specify IndexService to retrieve resources from\n" + 
        " -j                    Specify URL of job submission service to contact\n" +
        " -p                    Specify file that contains job preferences\n" +
		" -s                    Disable client file staging\n" + 
		" -cc                   Specify non-default client config file\n" +	
		" -cs                   Specify non-default client security descriptor file\n" + 
        "\n"; 

    private static final String usageDesc =
        "ARGUMENTS: \n" +
        "         [options] -f <job description file name>\n" +
        "with\n" +
        "         [options]       [-i <indexService]+ \n" +
        "                         [-p <job preferences file name>]\n" +
        "                         [-j <job submission service> (default is localhost)]\n" +
		"                         [-cc <client config file>\n" + 
		"                         [-cs <client security descriptor file>\n" + 
		"                         [-s]\n" + 
        "\n";

    public SubmitJob() throws Exception {
    	
        this.jobDescriptionFile = null; 
        this.jobPreferencesFile = null; 
        // Using resolved address instead of 127.0.0.1 which gives 
        // some problems with authentication as certificates differs. 
        InetAddress address = InetAddress.getLocalHost();     
        this.host = address.getCanonicalHostName(); 
        this.clientHost = host; 
        
        this.URI = ServiceURL.getURL(
        		host, JobSubmissionConstants.SERVICE_NAME).
					toString(); 
        
        this.indexServers = null; 
        
        this.clientStagesInput = true; 
        
        // Set default value for client config file 
    	String configFile = "/etc/jobsubmission/clientconfig.xml"; 
    	String globusLocation = ContainerConfig.getGlobusLocation(); 
    	this.clientConfigFile = globusLocation + configFile; 
    	
        // And for security descriptor file
        this.clientSecurityDescriptorFile = "etc/jobsubmission/client-security-config.xml";       
        
    }
    
    public void init() throws Exception {
    	
    	 // Read and deserialize config file 
        ClientConfiguration config = null;
        
        FileInputStream input = new FileInputStream(this.clientConfigFile); 
        InputSource inputSource = new InputSource(input); 
        config = (ClientConfiguration) ObjectDeserializer.deserialize(inputSource, 
        		ClientConfiguration.class); 
        
        this.converterClasses = config.getDescriptionConverter(); 
        this.stagingTargetResolverClasses = config.getStagingTargetResolver();                 
    	
    }

    /**
     * Parse user input
     * @param args
     * @throws Exception
     */
    public void processArguments(String args[]) throws Exception {

        List indexServersList = new Vector(); 

        if(args.length == 0) {
            System.out.println();
            System.out.println(usageDesc);
            System.out.println(optionsDesc);
            System.exit(0); 
        }

        boolean error = false; 
        for(int i = 0; i < args.length && !error; i++) {

            if (args[i].startsWith("-")) {

                if (args[i].equalsIgnoreCase("-help")) {
                    System.out.println();
                    System.out.println(usageDesc);
                    System.out.println(optionsDesc);
                    System.exit(0);
                }
                else if (args[i].equalsIgnoreCase("-usage")) {
                    System.out.println(usageDesc);
                    System.exit(0);
                }
                else if (args[i].equals("-cc")) {
                	i++; 
                	if(i == args.length) {
                		System.err.println("-c requires a client config file"); 
                		error = true; 
                	} 
                	else {
                		this.clientConfigFile = args[i]; 
                	}
                }
                else if (args[i].equals("-cs")) {
                	i++; 
                	if(i == args.length) {
                		System.err.println("-c requires a client security descriptor"); 
                		error = true; 
                	} 
                	else {
                		this.clientSecurityDescriptorFile = args[i]; 
                	}
                }                
                else if (args[i].equals("-f")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-f requires a job description file");
                        error = true; 
                    }
                    else
                        jobDescriptionFile = args[i]; 
                }
                else if (args[i].equals("-i")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-i requires a index service URI");
                        error = true; 
                    }
                    else 
                        indexServersList.add(args[i]); 
                }
                else if (args[i].equals("-j")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-j requires a job submission service URI");
                        error = true; 
                    }
                    else  {
                        host = args[i]; 
                
                        URI = ServiceURL.getURL(
                        		host, 
								JobSubmissionConstants.SERVICE_NAME).toString();
                    }
							
                }
                else if (args[i].equals("-p")) {
                    i++; 
                    if(i == args.length) {
                        System.err.println("-p requires a job preferences file");
                        error = true; 
                    }
                    else
                        jobPreferencesFile = args[i]; 
                }        
                else if (args[i].equals("-s")) {
                	System.out.println("File staging will be performed by "
                			+ "service, not by client"); 
                	this.clientStagesInput = false; 
                }
                else {
                    System.err.println("Unknown argument: " + args[i]);
                    error = true; 
                }
            }
            else {
                System.err.println("Unknown argument: " + args[i]);
                error = true; 
            }
        }
        if(error) {
            System.exit(1); 
        }

        int noIndexServers = indexServersList.size(); 

        if(noIndexServers > 0) {
            indexServers = new String[noIndexServers]; 
            for(int i = 0; i < noIndexServers; i++) 
                indexServers[i] = (String) indexServersList.get(i); 
        }
    }
       
     
    /**
     * Locate JobSubmissionService, convert job description
     * and submit it to service. Include any user preferences document
     * and/or non-default index server(s) in the job request. 
     * @throws Exception
     */
    public void submitJob() throws FileNotFoundException, Exception {

        System.out.println("Submitting job using:\n" + URI);

        System.out.println("Converting job description to JSDL");
        if(jobDescriptionFile == null) {
        	System.err.println("The -f flag (JobDescription) must be specified"); 
        	throw new 
				Exception("The -f flag (JobDescription) must be specified"); 
        }        
        
        JobDefinition_Type jobDesc = ClientUtils.convertJobDescription(
        		jobDescriptionFile, this.converterClasses);         
        
        System.out.println("Locating job submission service"); 
        // Locate service
        JobSubmissionServiceAddressingLocator locator = 
            new JobSubmissionServiceAddressingLocator(); 
                        
        JobSubmission jobsubmissionservice = null; 
        
        try { 
        	EndpointReferenceType endpoint = new EndpointReferenceType();
        	endpoint.setAddress(new Address(URI));
        	jobsubmissionservice = locator.getJobSubmissionPort(endpoint);        
        } catch (Exception e) {
        	System.err.println("Can not contact job submission service at " + URI); 
        	throw e; 
        }
                                
        // Set descriptor on Stub 
        ((Stub)jobsubmissionservice)._setProperty(
        		Constants.CLIENT_DESCRIPTOR_FILE, 
				this.clientSecurityDescriptorFile);
                
        // Build job request message
        JobRequest requestParameters = new JobRequest(); 
        // Index servers
        requestParameters.setIndexServerURIs(this.indexServers); 
        // JobDescription
        requestParameters.setJobDescription(jobDesc); 
        // Jobpreferences file 
        JobPreferencesType preferences = ClientUtils.getJobPreferences(this.jobPreferencesFile); 
        if(preferences != null) {
        	requestParameters.setJobPreferences(preferences); 
        }
        // ClientHost       
        requestParameters.setClientURL(this.clientHost);
        // Indicate whether client stages job files. 
        requestParameters.setClientStagesFiles(new Boolean(this.clientStagesInput)); 
        
        System.out.println("Submitting job..."); 
        String jobID = null; 
      
        // Perform actual submission
        try {
        	jobID = jobsubmissionservice.requestJob(requestParameters); 
        }  
        catch (InvalidJobDescriptionFaultType f) {
        	System.err.println("Job request failed, invalid job description:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch(NoAptResourcesFoundFaultType f) {
        	System.err.println("Job request failed"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.out.println(helper.getDescriptionAsString()); 
        } catch(JobRequestFaultType f) {
        	System.err.println("Job request failed:"); 
        	FaultHelper helper = new FaultHelper(f); 
        	System.err.println(helper.getDescriptionAsString()); 
        } catch (BaseFaultType f) {        	
        	System.err.println("General fault: "); 
        	FaultHelper helper = new FaultHelper(f);         	        	        	
        	System.err.println(helper.getDescriptionAsString());         	
        } catch(AxisFault f) {
        	System.err.println("Can not contact job submission service: " + URI); 
        	System.err.println("Reason: " + f.getFaultReason()); 
        } 
        
        if(jobID != null) {
        	try { 
        		if(this.clientStagesInput) {
        			System.out.println("Staging input files..."); 
        			if(ClientUtils.stageJobInput(jobID, 
        					jobDesc, 
							this.stagingTargetResolverClasses, 
							this.clientHost)) {
        				System.out.println("... done"); 
        			}
        			else {
        				System.out.println("... no input files to stage"); 
        			}
        			System.out.println("Job submitted with jobid " 
        					+ jobID); 
        		}
        	} catch (Exception e) {
        		System.err.println("Error staging input files: " + e.getMessage()); 
        		e.printStackTrace(); 
        		ClientUtils.cancelJob(jobID, 
        				this.stagingTargetResolverClasses); 
        	}                	
        }
    }
   

    public static void main(String args[]) {

        try {

        		System.out.println("Submitting job......."); 
            SubmitJob s = new SubmitJob(); 
            s.processArguments(args); 
            s.init(); 
            s.submitJob(); 

        } catch (FileNotFoundException e) {
        	System.err.println(e.getMessage()); 
        }
        catch (Exception e) {
            System.out.println("Error submitting job");
        }
    }
}


